/////////////////////////////////////////////////////////////////////////////
///
/// @file mapgraph.h
///
/// Computation of the graph representation of a combinatorial map.
/// This file contains the definition of a function for the computation
/// of a directed graph which represents a combinatorial map
/// restricted to a given set of cubes.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2008 by Pawel Pilarczyk.
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on February 12, 2007. Last revision: February 29, 2008.


#ifndef _CMGRAPHS_MAPGRAPH_H_
#define _CMGRAPHS_MAPGRAPH_H_


// include selected header files from the CHomP library
#include "chomp/system/textfile.h"
#include "chomp/cubes/pointset.h"
#include "chomp/cubes/cube.h"
#include "chomp/struct/digraph.h"

// include local header files
#include "config.h"
#include "typedefs.h"


// --------------------------------------------------
// -------------- compute cubical map ---------------
// --------------------------------------------------

// Computes the cubical map on 'X' and fills out the graph 'g'.
template <class typeCubes, class typeCubMap>
void computeMapGraph (const typeCubes &X, chomp::homology::diGraph<> &g,
	const typeCubMap &theCubMap)
{
	// remember the dimension of the phase space boxes
	int dim = X. size () ? X [0]. dim () : -1;

	// compute the images of all the cubes
	int nX = X. size ();
	for (int cur = 0; cur < nX; ++ cur)
	{
		// create the empty image
		g. addVertex ();

		// compute the image of the cube
		typeCubes img;
		theCubMap (X [cur], img);

		// add the corresponding edges to the graph
		int nImg = img. size ();
		for (int i = 0; i < nImg; ++ i)
		{
			int number = X. getnumber (img [i]);
			if (number < 0)
				continue;
			g. addEdge (number);
		}

		if ((dim > 3) && !(cur % 1000))
			chomp::homology::scon << std::setw (8) <<
				(cur / 1000) << "k\b\b\b\b\b\b\b\b\b";
	}

	return;
} /* computeMapGraph */


#endif // _CMGRAPHS_MAPGRAPH_H_


