/////////////////////////////////////////////////////////////////////////////
///
/// @file popmodel.h
///
/// The nonlinear Leslie population model map.
///
/// This map defines the nonlinear density dependent overcompensatory
/// population model, as described in the paper
/// by I. Ugarcovici and H. Weiss, "Chaotic dynamics of a nonlinear
/// density dependent population model", Nonlinearity 17 (2004), 1689-1711.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2008 by Pawel Pilarczyk.
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on February 9, 2008. Last revision: February 21, 2008.


#ifndef _CMGRAPHS_POPMODEL_H_
#define _CMGRAPHS_POPMODEL_H_

// the generic class for the type of a map
#include "maptype.h"

// the interval arithmetic package from the CAPD library
#include "interval/doubleInterval.h"


// --------------------------------------------------
// -------- the Leslie population model map ---------
// --------------------------------------------------

/// This class defines a map for the nonlinear density dependent
/// overcompensatory Leslie population model.
/// The interpretation of parameters depends on the dimension.
class MapPopModel: public MapType
{
public:
	/// Computes the image of a box whose left and right coordinates
	/// are given. Fills in the images with the left and right
	/// coordinates of the image box.
	void compute (const double *xleft, const double *xright,
		double *yleft, double *yright, int dim) const;

}; /* class MapPopModel */

// --------------------------------------------------

inline void MapPopModel::compute (const double *xleft, const double *xright,
	double *yleft, double *yright, int dim) const
{
	// an interval enclosure of the number e
	const interval e (2.7182818284590452, 2.7182818284590454);

	// compute the interval for the first coordinate of the result
	interval first = 0;
	interval sum = 0;
	for (int i = 0; i < dim; ++ i)
	{
		interval x = interval (xleft [i], xright [i]);
		first += interval (getLeftParam (i), getRightParam (i)) * x;
		sum += x;
	}
	first *= exp (-interval (getLeftParam (2 * dim - 1),
		getRightParam (2 * dim - 1)) * sum);

	// compute the remainder of the result
	// starting with the last coordinate
	for (int i = dim - 1; i > 0; -- i)
	{
		interval result = interval (getLeftParam (dim + i - 1),
			getRightParam (dim + i - 1)) *
			interval (xleft [i - 1], xright [i - 1]);
		yleft [i] = result. leftBound ();
		yright [i] = result. rightBound ();
	}

	// save the first coordinate of the result
	yleft [0] = first. leftBound ();
	yright [0] = first. rightBound ();

	// swich the rounding direction to the nearest representable number
	capd::rounding::DoubleRounding::roundNearest ();
	return;
} /* MapPopModel::compute */


#endif // _CMGRAPHS_POPMODEL_H_

